@@ -13,13 +13,13 @@
 
 package IO::Socket::SSL;
 
-our $VERSION = '2.000';
+our $VERSION = '2.004_1';
 
 use IO::Socket;
 use Net::SSLeay 1.46;
 use IO::Socket::SSL::PublicSuffix;
 use Exporter ();
-use Errno qw( EAGAIN ETIMEDOUT );
+use Errno qw( EWOULDBLOCK ETIMEDOUT EINTR );
 use Carp;
 use strict;
 
@@ -217,13 +217,23 @@ for(qw( SSLv2 SSLv3 TLSv1 TLSv1_1 TLSv11:TLSv1_1 TLSv1_2 TLSv12:TLSv1_2 )) {
     $SSL_OP_NO{$k} = eval { no strict 'refs'; &$sub } || 0;
 }
 
+# Make SSL_CTX_clear_options accessible through SSL_CTX_ctrl unless it is
+# already implemented in Net::SSLeay
+if (!defined &Net::SSLeay::CTX_clear_options) {
+    *Net::SSLeay::CTX_clear_options = sub {
+	my ($ctx,$opt) = @_;
+	# 77 = SSL_CTRL_CLEAR_OPTIONS
+	Net::SSLeay::CTX_ctrl($ctx,77,$opt,0);
+    };
+}
+
 our $DEBUG;
 use vars qw(@ISA $SSL_ERROR @EXPORT);
 
 {
     # These constants will be used in $! at return from SSL_connect,
     # SSL_accept, _generic_(read|write), thus notifying the caller
-    # the usual way of problems. Like with EAGAIN, EINPROGRESS..
+    # the usual way of problems. Like with EWOULDBLOCK, EINPROGRESS..
     # these are especially important for non-blocking sockets
 
     my $x = Net::SSLeay::ERROR_WANT_READ();
@@ -489,7 +499,7 @@ sub configure_SSL {
 
     # create context
     # this will fill in defaults in $arg_hash
-    $ctx ||= IO::Socket::SSL::SSL_Context->new($arg_hash);
+    $ctx ||= IO::Socket::SSL::SSL_Context->new($arg_hash) || return;
 
     ${*$self}{'_SSL_arguments'} = $arg_hash;
     ${*$self}{'_SSL_ctx'} = $ctx;
@@ -509,7 +519,7 @@ sub _skip_rw_error {
     } else {
 	return $err;
     }
-    $! ||= EAGAIN;
+    $! ||= EWOULDBLOCK;
     ${*$self}{'_SSL_last_err'} = $SSL_ERROR if ref($self);
     Net::SSLeay::ERR_clear_error();
     return 0;
@@ -1009,7 +1019,7 @@ sub _generic_write {
     }
     if ( !defined($written) ) {
 	if ( my $err = $self->_skip_rw_error( $ssl,-1 )) {
-	    $self->error("SSL write error");
+	    $self->error("SSL write error ($err)");
 	}
 	return;
     }
@@ -1064,9 +1074,9 @@ sub readline {
 	while (1) {
 	    my $rv = $self->sysread($buf,2**16,length($buf));
 	    if ( ! defined $rv ) {
-		next if $!{EINTR};                     # retry
-		last if $!{EAGAIN} || $!{EWOULDBLOCK}; # use everything so far
-		return;                                # return error
+		next if $! == EINTR;       # retry
+		last if $! == EWOULDBLOCK; # use everything so far
+		return;                    # return error
 	    } elsif ( ! $rv ) {
 		last
 	    }
@@ -1094,9 +1104,9 @@ sub readline {
 	while ( $size>length($buf)) {
 	    my $rv = $self->sysread($buf,$size-length($buf),length($buf));
 	    if ( ! defined $rv ) {
-		next if $!{EINTR};                     # retry
-		last if $!{EAGAIN} || $!{EWOULDBLOCK}; # use everything so far
-		return;                                # return error
+		next if $! == EINTR;       # retry
+		last if $! == EWOULDBLOCK; # use everything so far
+		return;                    # return error
 	    } elsif ( ! $rv ) {
 		last
 	    }
@@ -1115,7 +1125,7 @@ sub readline {
 	# wait until we have more data or eof
 	my $poke = Net::SSLeay::peek($ssl,1);
 	if ( ! defined $poke or $poke eq '' ) {
-	    next if $!{EINTR};
+	    next if $! == EINTR;
 	}
 
 	my $skip = 0;
@@ -1153,7 +1163,7 @@ sub readline {
 		$skip -= length($p);
 		next;
 	    }
-	    $!{EINTR} or last;
+	    $! == EINTR or last;
 	}
 
 	if ( $eod and ( $delim1 eq '' or $eod < length($buf))) {
@@ -1568,8 +1578,10 @@ if ( defined &Net::SSLeay::get_peer_cert_chain
 	if ( CAN_IPV6 and $identity =~m{:} ) {
 	    # no IPv4 or hostname have ':'  in it, try IPv6.
 	    $ipn = inet_pton(AF_INET6,$identity) or return; # invalid name
-	} elsif ( $identity =~m{^\d[\d\.]*\.\d+$} ) {
-	    $ipn = inet_aton($identity) or return; # invalid name
+	} elsif ( my @ip = $identity =~m{^(\d+)(?:\.(\d+)\.(\d+)\.(\d+)|[\d\.]*)$} ) {
+	    # check for invalid IP/hostname
+	    return if 4 != @ip or 4 != grep { defined($_) && $_<256 } @ip; 
+	    $ipn = pack("CCCC",@ip);
 	} else {
 	    # assume hostname, check for umlauts etc
 	    if ( $identity =~m{[^a-zA-Z0-9_.\-]} ) {
@@ -2091,6 +2103,8 @@ WARN
     }
 
     my $ssl_op = Net::SSLeay::OP_ALL();
+    $ssl_op |= &Net::SSLeay::OP_SINGLE_DH_USE;
+    $ssl_op |= &Net::SSLeay::OP_SINGLE_ECDH_USE if $can_ecdh;
 
     my $ver;
     for (split(/\s*:\s*/,$arg_hash->{SSL_version})) {
@@ -2145,6 +2159,13 @@ WARN
 	# SSL_OP_CIPHER_SERVER_PREFERENCE
 	$ssl_op |= 0x00400000 if $arg_hash->{SSL_honor_cipher_order};
 
+	if ($ver eq 'SSLv23' && !($ssl_op & $SSL_OP_NO{SSLv3})) {
+	    # At least LibreSSL disables SSLv3 by default in SSL_CTX_new.
+	    # If we really want SSL3.0 we need to explicitly allow it with
+	    # SSL_CTX_clear_options.
+	    Net::SSLeay::CTX_clear_options($ctx,$SSL_OP_NO{SSLv3});
+	}
+
 	Net::SSLeay::CTX_set_options($ctx,$ssl_op);
 
 	# if we don't set session_id_context if client certificate is expected
